# -*- coding: utf8 -*-

class Policy(object):
    """Decides which action to take."""
    def __init__(self):
        super(Policy, self).__init__()

    # def select_action(self, actions, values):
    #     return
